/*
 *  image.c -- part of stereographic2platecarree
 *
 *  Copyright (c) 2010 Marc Pegon <pe.marc@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "image.h"
#include "macros.h"
#include <assert.h>
#include <malloc.h>

void createUninitializedImage(Image *img, uint32_t width, uint32_t height)
{
    img->width = width;
    img->height = height;
    img->data = (uint32_t *)malloc(width * height * sizeof(uint32_t));

    if (img->data == NULL) {
        error_trace("Failed to create image.\n");
    }
}

void fillImage(Image img, uint32_t color)
{
    uint32_t *pixel = img.data;

    for (uint32_t i = 0; i < img.height; ++i) {
        for (uint32_t j = 0; j < img.width; ++j) {
            *pixel = color;
            ++pixel;
        }
    }
}

void createImage(Image *img, uint32_t width, uint32_t height, uint32_t color)
{
    createUninitializedImage(img, width, height);
    fillImage(*img, color);
}

void deleteImage(Image img)
{
    free(img.data);
}

uint32_t igetPixel(Image img, uint32_t x, uint32_t y)
{
    return img.data[y * img.width + x];
}

enum {LEFT = 0, UP, RIGHT, DOWN, UPLEFT, UPRIGHT, DOWNRIGHT, DOWNLEFT, CENTER};

int getRegion(Image img, double x, double y)
{
    int region;

    if (x < -1) {
        if (y < -1)
            region = UPLEFT;
        else if (y > img.height)
            region = DOWNLEFT;
        else
            region = LEFT;
    } else if (x > img.width) {
        if (y < -1)
            region = UPRIGHT;
        else if (y > img.height)
            region = DOWNRIGHT;
        else
            region = RIGHT;
    } else {
        if (y < -1)
            region = UP;
        else if (y > img.height)
            region = DOWN;
        else
            region = CENTER;
    }

    return region;
}

uint32_t mixPixels(uint32_t p1, double weight1, uint32_t p2, double weight2)
{
    uint32_t r1, g1, b1;
    uint32_t r2, g2, b2;

    r1 = GET8_R(p1);
    g1 = GET8_G(p1);
    b1 = GET8_B(p1);
    r2 = GET8_R(p2);
    g2 = GET8_G(p2);
    b2 = GET8_B(p2);

    uint32_t r = r1 * weight1 + r2 * weight2;
    uint32_t g = g1 * weight1 + g2 * weight2;
    uint32_t b = b1 * weight1 + b2 * weight2;

    return ((r << 16) + (g << 8) + b);
}

uint32_t fgetPixel(Image img, double x, double y)
{
    int region = getRegion(img, x, y);

    uint32_t pixels[4];
    double weights[4];
    double remainderX, remainderY;
    int signX, signY;
    uint32_t res;
    switch (region) {
    case LEFT:
        // use the pixel on the left of the image, at the given y coordinate
        remainderY = y - (uint32_t)y;
        if (remainderY == 0) {
            // y coordinate is integer : use igetPixel
            res = igetPixel(img, 0, (uint32_t)y);
        } else {
            // y coordinate isn't integer : mix pixels
            signY = (remainderY > 0) ? 1 : -1;
            pixels[0] = igetPixel(img, 0, (uint32_t)y);
            weights[0] = 1. - (double)signY * remainderY;
            pixels[1] = igetPixel(img, 0, (uint32_t)y + signY);
            weights[1] = (double)signY * remainderY;
            res = mixPixels(pixels[0], weights[0], pixels[1], weights[1]);
        }
        break;
    case UP:
        remainderX = x - (uint32_t)x;
        if (remainderX == 0) {
            res = igetPixel(img, (uint32_t)x, 0);
        } else {
            signX = (remainderX > 0) ? 1 : -1;
            pixels[0] = igetPixel(img, (uint32_t)x, 0);
            weights[0] = 1. - (double)signX * remainderX;
            pixels[1] = igetPixel(img, (uint32_t)x + signX, 0);
            weights[1] = (double)signX * remainderX;
            res = mixPixels(pixels[0], weights[0], pixels[1], weights[1]);
        }
        break;
    case RIGHT:
        remainderY = y - (uint32_t)y;
        if (remainderY == 0) {
            res = igetPixel(img, img.width - 1, (uint32_t)y);
        } else {
            signY = (remainderY > 0) ? 1 : -1;
            pixels[0] = igetPixel(img, img.width - 1, (uint32_t)y);
            weights[0] = 1. - (double)signY * remainderY;
            pixels[1] = igetPixel(img, img.width - 1, (uint32_t)y + signY);
            weights[1] = (double)signY * remainderY;
            res = mixPixels(pixels[0], weights[0], pixels[1], weights[1]);
        }
        break;
    case DOWN:
        remainderX = x - (uint32_t)x;
        if (remainderX == 0) {
            res = igetPixel(img, (uint32_t)x, img.height - 1);
        } else {
            signX = (remainderX > 0) ? 1 : -1;
            pixels[0] = igetPixel(img, (uint32_t)x, img.height - 1);
            weights[0] = 1. - (double)signX * remainderX;
            pixels[1] = igetPixel(img, (uint32_t)x + signX, img.height - 1);
            weights[1] = (double)signX * remainderX;
            res = mixPixels(pixels[0], weights[0], pixels[1], weights[1]);
        }
        break;
    case UPLEFT:
        res = igetPixel(img, 0, 0);
        break;
    case UPRIGHT:
        res = igetPixel(img, img.width - 1, 0);
        break;
    case DOWNLEFT:
        res = igetPixel(img, 0, img.height - 1);
        break;
    case DOWNRIGHT:
        res = igetPixel(img, img.width - 1, img.height - 1);
        break;
    case CENTER:
        // Handle the edges : doing this is like duplicating the edges
        if (y < 0) {
            y = 0;
        } else if (y > img.height - 1) {
            y = img.height - 1;
        }

        if (x < 0) {
            x = 0;
        } else if (x > img.height - 1) {
            x = img.height - 1;
        }

        remainderX = x - (uint32_t)x;
        remainderY = y - (uint32_t)y;
        if (remainderX == 0) {
            if (remainderY == 0) {
                // if x & y coordinates are integer : use igetPixel
                res = igetPixel(img, (uint32_t)x, (uint32_t)y);
            } else {
                // x coordinate is integer : we only mix pixels along y
                signY = (remainderY > 0) ? 1 : -1;
                pixels[0] = igetPixel(img, (uint32_t)x, (uint32_t)y);
                weights[0] = 1. - (double)signY * remainderY;
                pixels[1] = igetPixel(img, (uint32_t)x, (uint32_t)y + signY);
                weights[1] = (double)signY * remainderY;
                res = mixPixels(pixels[0], weights[0], pixels[1], weights[1]);
            }
        } else {
            if (remainderY == 0) {
                // y coordinate is integer : we only mix pixels along x
                signX = (remainderX > 0) ? 1 : -1;
                pixels[0] = igetPixel(img, (uint32_t)x, (uint32_t)y);
                weights[0] = 1. - (double)signX * remainderX;
                pixels[1] = igetPixel(img, (uint32_t)x + signX, (uint32_t)y);
                weights[1] = (double)signX * remainderX;
                res = mixPixels(pixels[0], weights[0], pixels[1], weights[1]);
            } else {
                // neither x nor y is integer : we mix pixels along x, then along y
                signX = (remainderX > 0) ? 1 : -1;
                signY = (remainderY > 0) ? 1 : -1;

                // Mix X
                pixels[2] = igetPixel(img, (uint32_t)x, (uint32_t)y);
                weights[2] = 1. - (double)signX * remainderX;
                pixels[3] = igetPixel(img, (uint32_t)x + signX, (uint32_t)y);
                weights[3] = (double)signX * remainderX;
                pixels[0] = mixPixels(pixels[2], weights[2], pixels[3], weights[3]);

                pixels[2] = igetPixel(img, (uint32_t)x, (uint32_t)y + signY);
                weights[2] = 1. - (double)signX * remainderX;
                pixels[3] = igetPixel(img, (uint32_t)x + signX, (uint32_t)y + signY);
                weights[3] = (double)signX * remainderX;
                pixels[1] = mixPixels(pixels[2], weights[2], pixels[3], weights[3]);

                // Mix Y
                weights[0] = 1. - (double)signY * remainderY;
                weights[1] = (double)signY * remainderY;
                res = mixPixels(pixels[0], weights[0], pixels[1], weights[1]);
            }
        }
        break;
    default:
        assert(0);
    }

    return res;
}

