/*
 *  main.c -- part of stereographic2platecarree
 *
 *  Copyright (c) 2010 Marc Pegon <pe.marc@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "ppm.h"
#include "image.h"
#include "projection.h"
#include "macros.h"
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>

void displayHelp()
{
    puts("\nConvert stereographic projection to plate carrée projection :\n  \
Usage : ./stereographic2platecarree [-h] [-q|-v] [-d] -i <Input> -o <OutPut> [-s <Size>]\n  \
-h : Prints this help.\n  \
-q : Quiet mode.\n  \
-v : Verbose mode.\n  \
-d : Debug mode.\n  \
-i <Input> : Specify the input image file (raw PPM only).  \
It MUST be a square image of a stereographic projection of  \
one half of the globe (correctly made, centered, etc).\n  \
-o <Output>: Specify the output image file (raw PPM).  \
If not specified, its size will be the same as input. \n  \
-s <Size> : Specify the size of the output image (it will  \
be a square image, so only one value is expected). \n  \
    ");
}

int main(int argc, char **argv)
{
    traceLevel = T_NORMAL;
    debug = 0;
    int help = 0;
    int dstSizeSpecified = 0;
    uint32_t dstSize = 0;
    char *srcFileName = NULL;
    char *dstFileName = NULL;
    int64_t tmp;
    int o;
    while ((o = getopt(argc, argv, "hqvds:i:o:")) != -1) {
        switch (o) {
        case 'h':
            help = 1;
            break;
        case 'q':
            traceLevel = T_QUIET;
            break;
        case 'v':
            traceLevel = T_VERBOSE;
            break;
        case 'd':
            debug = 1;
            break;
        case 's':
            sscanf(optarg, "%lld", &tmp);
            if (tmp < 0) {
                info_trace(T_QUIET, "Size must be positive.\n");
                return EXIT_FAILURE;
            } else {
                dstSizeSpecified = 1;
                dstSize = (uint32_t)tmp;
            }
            break;
        case 'i':
            srcFileName = optarg;
            break;
        case 'o':
            dstFileName = optarg;
            break;
        default:
            displayHelp();
            return EXIT_FAILURE;
            break;
        }
    }

    if (help) {
        displayHelp();
        return EXIT_SUCCESS;
    }

    if (srcFileName == NULL) {
        info_trace(T_QUIET, "Invalid use : no input file specified.\n");
        displayHelp();
        return EXIT_FAILURE;
    }

    if (dstFileName == NULL) {
        info_trace(T_QUIET, "Invalid use : no output file specified.\n");
        displayHelp();
        return EXIT_FAILURE;
    }

    Image src, dst;
    PPMImport(&src, srcFileName);

    if (src.width != src.height) {
        deleteImage(src);
        error_trace("Bad input image. Width and height must be equal.\n");
    }

    if (!dstSizeSpecified) {
        dstSize = src.width;
    }

    createUninitializedImage(&dst, dstSize, dstSize);
    stereographic2PlateCarree(dst, src);

    PPMExport(dst, dstFileName);

    deleteImage(src);
    deleteImage(dst);

    return EXIT_SUCCESS;
}

