/*
 *  ppm.c -- part of stereographic2platecarree
 *
 *  Copyright (c) 2010 Marc Pegon <pe.marc@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "ppm.h"
#include "macros.h"
#include <assert.h>
#include <string.h>
#include <ctype.h>

// Requires image is initialized
void PPMExport(Image image, char *fileName)
{
    FILE *file;

    info_trace(T_NORMAL, "Exporting PPM \'%s\'..\n", fileName);
    file = fopen(fileName, "w");
    if (!file) {
        open_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\t%s opened successfully.\n", fileName);
    }

    if (fprintf(file,"P6\n") < 0) {
        write_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tWrote magic number (P6).\n");
    }

    if (fprintf(file, "%d ", image.width) < 0) {
        write_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tWrote width (%d).\n", image.width);
    }
        
    if (fprintf(file, "%d\n", image.height) < 0) {
        write_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tWrote height (%d).\n", image.height);
    }
        
    if (fprintf(file, "%d\n", UINT8_MAX) < 0) {
        write_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tWrote Maxval (%d).\n", UINT8_MAX);
    }

    uint32_t *pixel = image.data;
    for (uint32_t i = 0 ; i < image.height ; ++i) {
        for (uint32_t j = 0 ; j < image.width ; ++j) {
            if (fprintf(file, "%c%c%c", GET8_R(*pixel), GET8_G(*pixel), GET8_B(*pixel)) < 0) {
                write_error(fileName);
            }
            ++pixel;
        }
    }

    info_trace(T_VERBOSE, "\tWrote data.\n");

    if (fclose(file)) {
        close_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tClosed file successfully\n");
    }

    info_trace(T_NORMAL, "Export done.\n");
}

void SkipWhiteSpaces(FILE *file)
{
    int c;

    while (!feof(file)) {
        c = fgetc(file);

        if (!isspace(c)) {
            fseek(file, -1, SEEK_CUR);
            break;
        }
    }
}

void SkipComments(FILE *file, char *fileName)
{
    char line[255];
    int c;

    do {
        SkipWhiteSpaces(file);
        c = fgetc(file);

        if (c == EOF) {
            read_error(fileName);
        } else if (c == '#') {
            char *lineRead = fgets(line, 255, file);
            if (lineRead == NULL) {
                read_error(fileName);
            }
        } else {
            // '#' not found
            fseek(file, -1, SEEK_CUR);
        }
    } while (c == '#');
}

// Requires *image exists but is uninitalized.
void PPMImport(Image *image, char *fileName)
{
    FILE *file;

    info_trace(T_NORMAL, "Importing PPM \'%s\'..\n", fileName);
    file = fopen(fileName, "r");
    if (!file) {
        open_error(fileName);
    }

    SkipComments(file, fileName);
    char magicNumber[3];
    if (fgets(magicNumber, 3, file) == NULL) {
        read_error(fileName);
    }

    SkipComments(file, fileName);
    if (strcmp(magicNumber, "P6") != 0) {
        error_trace("Unrecognized magic number : %s. Expected : P6\n", magicNumber);
    } else {
        info_trace(T_VERBOSE, "\tRead magic number (P6).\n");
    }
    
    uint32_t width, height;
    SkipComments(file, fileName);
    if (fscanf(file, "%u", &width) == EOF) {
        read_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tRead image width (%d).\n", width);
    }

    SkipComments(file, fileName);
    if (fscanf(file, "%u", &height) == EOF) {
        read_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tRead image height (%d).\n", height);
    }

    SkipComments(file, fileName);
    uint32_t maxVal;
    int bytesPerComponent;
    if (fscanf(file, "%u", &maxVal) == EOF) {
        read_error(fileName);
    } else {
        if (maxVal > UINT16_MAX) {
            error_trace("Maxval is too big : %u. Maximum is %u\n", maxVal, UINT16_MAX);
        } else if (maxVal > UINT8_MAX) {
            bytesPerComponent = 2;
            info_trace(T_NORMAL, "Warning: image needs 2 bytes per component. Quality will be reduced.\n");
        } else {
            bytesPerComponent = 1;
        }
        info_trace(T_VERBOSE, "\tRead image Maxval (%d).\n", maxVal);
    }

    int c = fgetc(file);
    if (c == EOF) {
        read_error(fileName);
    }

    if (!isspace(c)) {
        error_trace("Whitespace character expected. Found %c\n", c);
    }

    createUninitializedImage(image, width, height);
    uint32_t *pixel = image->data;
    int r, g, b;
    if (bytesPerComponent == 1) {
        for (uint32_t i = 0 ; i < image->height ; ++i) {
            for (uint32_t j = 0 ; j < image->width ; ++j) {
                r = fgetc(file);
                if (r == EOF || r > UINT8_MAX) {
                    read_error(fileName);
                }
                g = fgetc(file);
                if (g == EOF || g > UINT8_MAX) {
                    read_error(fileName);
                }
                b = fgetc(file);
                if (b == EOF || b > UINT8_MAX) {
                    read_error(fileName);
                }

                *pixel = (r << 16) + (g << 8) + b;
                ++pixel;
            }
        }
    } else if (bytesPerComponent == 2) {
        for (uint32_t i = 0 ; i < image->height ; ++i) {
            for (uint32_t j = 0 ; j < image->width ; ++j) {
                r = fgetc(file);
                if (r == EOF || r > UINT16_MAX) {
                    read_error(fileName);
                }
                g = fgetc(file);
                if (g == EOF || g > UINT16_MAX) {
                    read_error(fileName);
                }
                b = fgetc(file);
                if (b == EOF || b > UINT16_MAX) {
                    read_error(fileName);
                }

                r = (r * UINT8_MAX) / UINT16_MAX;
                g = (g * UINT8_MAX) / UINT16_MAX;
                b = (b * UINT8_MAX) / UINT16_MAX;
                *pixel = (r << 16) + (g << 8) + b;
                ++pixel;
            }
        }
    } else {
        assert(0);
    }

    info_trace(T_VERBOSE, "\tData successfully read.\n");

    if (fclose(file)) {
        close_error(fileName);
    } else {
        info_trace(T_VERBOSE, "\tClosed file successfully\n");
    }

    info_trace(T_NORMAL, "Import done.\n");
}
