/*
 *  projection.c -- part of stereographic2platecarree
 *
 *  Copyright (c) 2010 Marc Pegon <pe.marc@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "projection.h"
#include "macros.h"
#include <math.h>

fPoint tofPoint(double x, double y)
{
    fPoint res;
    // x <=> phi
    res.x = x;
    // y <=> theta
    res.y = y;

    return res;
}

fPoint math_PlateCarree2Spheric(fPoint p, double radius)
{
    return tofPoint(p.x * M_PI_2 / radius, p.y * M_PI_2 / radius);
}

fPoint math_Spheric2Stereographic(fPoint p, double radius)
{
    double phi = p.x;
    double theta = p.y;

    double cos_phi = cos(phi);
    double sin_phi = sin(phi);
    double cos_theta = cos(theta);
    double sin_theta = sin(theta);

    double denom = 1. + sin_theta * sin_phi;

    double x = radius * (1 - cos_phi * sin_theta / denom);
    double y = radius * (1 - cos_theta / denom);

    return tofPoint(x, y);
}

void stereographic2PlateCarree(Image dst, Image src)
{
    info_trace(T_NORMAL, "Converting projection..\n");

    if (dst.width != dst.height) {
        error_trace("Dst should be a square image.\n");
    }
    if (src.width != src.height) {
        error_trace("Src should be a square image.\n");
    }

    double srcRadius = (double)src.width / 2.;
    double dstRadius = (double)dst.width / 2.;

    uint32_t *dstPixel = dst.data;
    uint32_t color;
    fPoint pDst, pSrc, pSpheric;
    for (uint32_t i = 0; i < dst.height; ++i) {
        for (uint32_t j = 0; j < dst.width; ++j) {
            pDst = tofPoint(j, i);
            pSpheric = math_PlateCarree2Spheric(pDst, dstRadius);
            pSrc = math_Spheric2Stereographic(pSpheric, srcRadius);
            color = fgetPixel(src, pSrc.x, pSrc.y);
            *dstPixel = color;
            ++dstPixel;
        }
    }

    info_trace(T_NORMAL, "Conversion done.\n");
}
